import xml.etree.ElementTree as ET
import csv
import sys
from collections import defaultdict
import xml.dom.minidom
from datetime import datetime, timedelta

def insert_target_comment(input_ptr_path,
                          output_ptr_path=False,
                          target='JUPITER',
                          verbose=True):
    """
    Insert a target comment into the metadata of each block in an XML file.

    This function reads an XML file and appends a comment indicating the specified target 
    into the metadata of each block element.

    Parameters:
    input_ptr_path (str): The path to the input XML file.
    output_ptr_path (str, optional): The path for the output XML file. 
                                      Defaults to the input file name with an added '.output' extension.
    target (str): The target name to insert into the comments. Defaults to 'JUPITER'.
    verbose (bool): If True, prints the modified XML. Defaults to True.

    Returns:
    str: The modified XML as a string.
    """
    
    if not output_ptr_path:
        output_ptr_path = input_ptr_path + '.output'

    # Read the XML content from the file
    with open(input_ptr_path, 'r') as file:
        xml_input = file.read()

    root = ET.fromstring(xml_input)

    # Iterate through all 'block' elements and check for the target
    for block in root.iter('block'):
        attitude = block.find('attitude')
        if attitude is not None:
            metadata = block.find('metadata')
            if metadata is not None:
                comment_element = ET.Element('comment')
                comment_element.text = f'TARGET={target}'
                metadata.append(comment_element)

    # Convert the modified XML back to string
    modified_xml = ET.tostring(root, encoding='unicode')
    # Write the modified XML to a new file
    with open(output_ptr_path, 'w') as file:
        file.write(modified_xml)

    if verbose:
        print(modified_xml)

    return modified_xml


def set_obs_id(input_ptr_path,
               output_ptr_path=False,
               verbose=False):
    """
    Set or update the OBS_ID in the metadata of OBS blocks in an XML file.

    This function reads an XML file, increments the OBS_ID for each OBS block, 
    and appends or updates the corresponding OBS_ID comment in the metadata.

    Parameters:
    input_ptr_path (str): The path to the input XML file.
    output_ptr_path (str, optional): The path for the output XML file. 
                                      Defaults to the input file name with an added '.output' extension.
    verbose (bool): If True, prints the modified XML. Defaults to False.

    Returns:
    str: The modified XML as a string.
    """
    
    obs_id = 0

    if not output_ptr_path:
        output_ptr_path = input_ptr_path + '.output'

    # Read the XML content from the file
    with open(input_ptr_path, 'r') as file:
        xml_input = file.read()

    root = ET.fromstring(xml_input)

    # Iterate through all 'block' elements and check for the target
    for block in root.iter('block'):
        if block.attrib.get('ref') == 'OBS':
            obs_id += 1
            metadata = block.find('metadata')
            if metadata is not None:
                obs_id_present = False
                # Iterate through existing comments to find and update OBS_ID
                for comment in metadata.findall(".//comment"):
                    if 'OBS_ID' in comment.text:
                        comment.text = f' OBS_ID={obs_id:03d} '
                        obs_id_present = True

                # If OBS_ID comment element is not present, add it
                if not obs_id_present:
                    comment_element = ET.Element('comment')
                    comment_element.text = f' OBS_ID={obs_id:03d} '
                    metadata.append(comment_element)

    #  <comment> Track Power Optimised Jupiter </comment>
    #  <comment> PRIME=UVS </comment>
    #  <comment> OBS_ID=002 </comment>
    #  <comment> OBS_NAME=UVS_JUP_AP_SCAN_MAP </comment>
    #  <comment> TARGET=JUPITER </comment>

    # Convert the modified XML back to string
    modified_xml = ET.tostring(root, encoding='unicode')
    # Write the modified XML to a new file
    with open(output_ptr_path, 'w') as file:
        file.write(modified_xml)

    if verbose:
        print(modified_xml)

    return modified_xml


def check_obs_metadata(xml_file, output_ptr_path=False):
    """
    Check for missing metadata fields in OBS blocks of an XML file and update the XML accordingly.

    This function reads an XML file, checks each 'block' element with a reference of 'OBS'
    for required metadata fields. If any required fields are missing, it adds placeholder comments
    and generates a report of the missing fields.

    Parameters:
    xml_file (str): The path to the input XML file.
    output_ptr_path (str, optional): The path for the output XML file with updates.
                                      Defaults to the input XML file name with an added '.output' extension.

    Returns:
    str: The modified XML as a string.
    """

    # Read the XML content from the file
    tree = ET.parse(xml_file)
    root = tree.getroot()

    if not output_ptr_path:
        output_ptr_path = xml_file + '.output'

    # List to store missing fields for each OBS
    missing_fields_report = []

    for block in root.iter('block'):
        if block.attrib.get('ref') == 'OBS':
            attitude = block.find('attitude')
            start_time =  block.find('startTime')
            end_time = block.find('endTime')
            try:
                start_time = start_time.text.strip()
            except:
                start_time = ''
            try:
                end_time = end_time.text.strip()
            except:
                end_time = ''
            if attitude is not None:
                metadata = block.find('metadata')
                if metadata is not None:
                    obs_info = {'OBS_NAME': '', 'OBS_ID': '',
                                'BLOCK_START': start_time, 'BLOCK_END': end_time}
                    obs_id_found = False

                    # Check for required fields and add if missing
                    for comment in metadata.iter('comment'):
                        for field in ['PRIME', 'OBS_NAME', 'TARGET', 'OBS_ID']:
                            if field in comment.text:
                                value = comment.text.split('=')[1].strip()
                                obs_info[field] = value

                    for field in ['PRIME', 'OBS_NAME', 'TARGET', 'OBS_ID']:
                        if field not in obs_info:
                            obs_info[field] = 'MISSING'
                            # Add the missing field with a placeholder value
                            comment_element = ET.Element('comment')
                            comment_element.text = f' {field}=PLACEHOLDER '
                            metadata.append(comment_element)

                            # Add the OBS information to the report
                            missing_fields_report.append(obs_info)
                if not obs_info['OBS_NAME']: obs_info['OBS_NAME'] = '???'
                if not obs_info['OBS_ID']: obs_info['OBS_ID'] = '???'

    # Convert the modified XML back to string
    modified_xml = ET.tostring(root, encoding='unicode')

    print("\nMissing Fields Report:")
    for obs_info in missing_fields_report:
        print(f"OBS_NAME: {obs_info['OBS_NAME']}, OBS_ID: {obs_info['OBS_ID']}")
        print(f"BLOCK_START: {obs_info['BLOCK_START']}, BLOCK_END: {obs_info['BLOCK_END']}")
        for field, status in obs_info.items():
            if status == 'MISSING':
                print(f"  - {field} is missing")
        print()

    # Write the modified XML to a new file
    with open(output_ptr_path, 'w') as file:
        file.write(modified_xml)

    return modified_xml


def convert_ptr_to_opl(xml_file: str, output_ptr_path: str = None, output_split: bool = False):
    """
    Export observations from an XML file into CSV format.

    This function reads an XML file containing observation data and 
    exports it either into separate CSV files for each designer or 
    into a single CSV file, based on the parameters provided.

    Parameters:
    xml_file (str): The path to the input XML file.
    output_ptr_path (str, optional): The path for the output CSV file(s). 
                                      Defaults to the input XML file name with an added '.output' extension.
    output_split (bool): If True, generates separate CSV files for each designer. 
                         If False, generates a single output file.

    Returns:
    None
    """

    # Read the XML content from the file
    tree = ET.parse(xml_file)
    root = tree.getroot()

    # Determine the output file path
    if not output_ptr_path:
        output_ptr_path = xml_file + '.output'

    # Determine whether to output separately or all together
    separate_files = output_split  
    
    if separate_files:
        # Prepare a dictionary to hold file handles for each designer
        designer_files = defaultdict(list)
    else:
        # Prepare to write to a single CSV file
        single_file_name = output_ptr_path
        single_file_handle = open(single_file_name, mode='w', newline='')
        single_writer = csv.writer(single_file_handle)

    for block in root.findall('.//block'):
        start_time = block.find('startTime')
        end_time = block.find('endTime')

        if start_time is not None and end_time is not None:
            start_time = start_time.text.strip()
            end_time = end_time.text.strip()

            attitude = block.find('attitude')
            metadata = block.find('metadata')

            if attitude is not None and metadata is not None:
                observations_elements = metadata.findall('observations')
                observations_element = metadata.find('observations')

                if observations_element is None:
                    print("WARNING: No observations found in the PTR for startTime " + start_time)

                elif len(observations_elements) == 1:
                    for observations_element in observations_elements:
                        designer = observations_element.get('designer')
                        if designer is None:
                            print("WARNING: Missing 'designer' attribute in observations. It is mandatory.")
                            continue

                        if separate_files:
                            # Initialize a CSV file for this designer if not already done
                            if designer not in designer_files:
                                file_name = f'ptr_output_{designer}.csv'
                                file_handle = open(file_name, mode='w', newline='')
                                writer = csv.writer(file_handle)
                                designer_files[designer] = (file_handle, writer)
                                sys.stdout.write(f'OPL structure saved in: {file_name}\n')

                            current_writer = designer_files[designer][1]
                        else:
                            current_writer = single_writer

                        # Extract observations
                        observation_list = observations_element.findall('observation')

                        for observation in observation_list:
                            obs_type = observation.find('type')
                            definition = observation.find('definition')
                            unit = observation.find('unit')
                            target = observation.find('target')
                            start_time_obs = observation.find('startTime')
                            end_time_obs = observation.find('endTime')

                            if None in [obs_type, definition, unit, target, start_time_obs, end_time_obs]:
                                print("WARNING: One or more fields are missing in an observation.")

                            obs_type_upper = obs_type.text.strip().upper() if obs_type is not None else "UNKNOWN"
                            definition = definition.text.strip() if definition is not None else "UNKNOWN"
                            unit = unit.text.strip() if unit is not None else "UNKNOWN"
                            target = target.text.strip() if target is not None else "UNKNOWN"
                            start_time_obs = start_time_obs.text.strip() + 'Z' if start_time_obs is not None else "UNKNOWN"
                            end_time_obs = end_time_obs.text.strip() + 'Z' if end_time_obs is not None else "UNKNOWN"
                            obs_type_output = f"{unit}_{obs_type_upper}_OBSERVATION"

                            # Write to the corresponding writer
                            current_writer.writerow([
                                obs_type_output,
                                start_time_obs,
                                end_time_obs,
                                definition,
                                unit
                            ])
                else:
                    print("WARNING: Block skipped, more than one observations node occurrence in the PTR for startTime " + start_time)


    # Close all file handles
    if separate_files:
        for file_handle, _ in designer_files.values():
            file_handle.close()
    else:
        single_file_handle.close()
        sys.stdout.write(f'OPL structure saved in: {single_file_name}\n')


# Function to parse the OPL CSV file and return observations as a list of dictionaries
def parse_opl(opl_file):
    observations = []
    with open(opl_file, "r") as csvfile:
        reader = csv.reader(csvfile)
        unique_rows = set(tuple(row) for row in reader)  # Remove duplicate rows
        for row in unique_rows:
            if len(row) != 5:
                continue  # Skip rows with incorrect format
            try:
                start_time = datetime.strptime(row[1], "%Y-%m-%dT%H:%M:%S.%fZ")  # Try with fractional seconds and Z
            except ValueError:
                start_time = datetime.strptime(row[1], "%Y-%m-%dT%H:%M:%S")  # Fallback to naive datetime

            try:
                end_time = datetime.strptime(row[2], "%Y-%m-%dT%H:%M:%S.%fZ")  # Try with fractional seconds and Z
            except ValueError:
                end_time = datetime.strptime(row[2], "%Y-%m-%dT%H:%M:%S")  # Fallback to naive datetime

            observation = {
                "type": row[0].split("_")[1] if "OBSERVATION" in row[0] else "DESIGNER",
                "start_time": start_time,
                "end_time": end_time,
                "name": row[3] if "OBSERVATION" in row[0] else row[0],
                "unit": row[4] if "OBSERVATION" in row[0] else "SOC"
            }

            observations.append(observation)

            if observation["unit"] == "SOC":
                soc_prime_observation = {
                "type": "PRIME",
                "start_time": start_time,
                "end_time": end_time,
                "name": row[3] if "OBSERVATION" in row[0] else row[0],
                "unit": row[4] if "OBSERVATION" in row[0] else "SOC"
                }
                observations.append(soc_prime_observation)

    return observations

# Function to prettify XML without extra empty lines
def prettify(element):
    rough_string = ET.tostring(element, encoding="utf-8")
    parsed = xml.dom.minidom.parseString(rough_string)
    pretty_xml = parsed.toprettyxml(indent="  ")
    lines = [line for line in pretty_xml.splitlines() if line.strip()]  # Remove empty lines
    return "\n".join(lines)

# Function to format delta time in (+/-)hh:mm:ss format
def format_delta(delta_seconds):
    delta = timedelta(seconds=abs(delta_seconds))
    hours, remainder = divmod(delta.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    sign = "+" if delta_seconds >= 0 else "-"
    return f"{sign}{hours:02}:{minutes:02}:{seconds:02}"

# Function to update PTR XML with observations from the OPL
def update_ptr_with_opl(ptr_file, opl_file, output_file):
    # Parse the OPL file
    opl_observations = parse_opl(opl_file)

    # Parse the PTR XML file
    tree = ET.parse(ptr_file)
    root = tree.getroot()

    # Iterate over PTR blocks and add observations
    for block in root.findall(".//block"):
        block_ref = block.get("ref")  # Get the reference of the block
        if block_ref != "OBS":  # Skip blocks that are not OBS
            continue

        block_start_time = block.find("startTime").text
        block_end_time = block.find("endTime").text

        # Parse block start and end times with both possible formats
        try:
            block_start = datetime.strptime(block_start_time, "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            block_start = datetime.strptime(block_start_time, "%Y-%m-%dT%H:%M:%S")

        try:
            block_end = datetime.strptime(block_end_time, "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            block_end = datetime.strptime(block_end_time, "%Y-%m-%dT%H:%M:%S")

        # Determine the designer for this block
        designer_instrument = "SOC"  # Default to SOC
        for obs in opl_observations:
            if obs["type"] == "DESIGNER" and block_start <= obs["start_time"] < block_end:
                designer_instrument = obs["unit"]
                break

        # Clean up existing metadata
        metadata = block.find("metadata")
        if metadata is not None:
            # Remove <comment> elements
            for comment in metadata.findall("comment"):
                metadata.remove(comment)

            # Remove <planning> elements
            planning = metadata.find("planning")
            if planning is not None:
                metadata.remove(planning)

        # Create the <observations> element
        planning = ET.SubElement(metadata, "planning")
        observations_element = ET.SubElement(planning, "observations", designer=designer_instrument)

        # Add PRIME and RIDER observations to the timeline
        for obs in opl_observations:
            obs_start = obs["start_time"]
            obs_end = obs["end_time"]

            # Check observation inclusion rules
            if obs_end == block_start or obs_start == block_end:
                continue  # Skip observations ending at block start or starting at block end

            # Include only PRIME and RIDER observations in the timeline
            if obs["type"] not in ["PRIME", "RIDER"]:
                continue

            if block_start <= obs_start < block_end or block_start < obs_end <= block_end or (obs_start <= block_start and obs_end >= block_end):
                observation_element = ET.SubElement(observations_element, "observation")

                ET.SubElement(observation_element, "type").text = obs["type"]
                ET.SubElement(observation_element, "definition").text = obs["name"]
                ET.SubElement(observation_element, "unit").text = obs["unit"]

                # Calculate startDelta and endDelta based on block boundaries
                start_delta = max(0, (obs_start - block_start).total_seconds())
                end_delta = min((obs_end - block_end).total_seconds(), 0)

                ET.SubElement(observation_element, "startDelta").text = format_delta(start_delta)
                ET.SubElement(observation_element, "endDelta").text = format_delta(end_delta)
                ET.SubElement(observation_element, "startTime").text = obs_start.strftime("%Y-%m-%dT%H:%M:%S")
                ET.SubElement(observation_element, "endTime").text = obs_end.strftime("%Y-%m-%dT%H:%M:%S")

    # Prettify and write the updated PTR back to a new file
    with open(output_file, "w", encoding="utf-8") as output:
        output.write(prettify(root))


